from fastapi import FastAPI, APIRouter
from .openai import add_routes

def create_app(title='svllm', description='svllm api', prefix='/v1', root_path='', *args, **kwargs) -> FastAPI:
    app = FastAPI(title=title, description=description, root_path=root_path, *args, **kwargs)
    app.include_router(add_routes(APIRouter(prefix=prefix)))
    return app



from fastapi.testclient import TestClient
import json

def run_sync_complete_tests(client: TestClient):
    print('Running sync complete tests...')
    response = client.post('/v1/completions', json={
        'model': 'svllm-complete',
        'prompt': 'Hello',
        'stream': False
    })
    assert response.status_code == 200
    assert 'choices' in response.json()
    assert isinstance(response.json()['choices'], list)
    assert len(response.json()['choices']) == 1
    assert 'text' in response.json()['choices'][0]
    assert isinstance(response.json()['choices'][0]['text'], str)
    assert response.json()['choices'][0]['text'].startswith('The Zen of Python')

def run_stream_complete_tests(client: TestClient):
    print('Running stream complete tests...')
    response = client.post('/v1/completions', json={
        'model': 'svllm-complete',
        'prompt': 'Hello',
        'stream': True
    })

    assert response.status_code == 200

    content = ''
    for chunk in response.iter_lines():
        chunk = chunk.strip()
        if chunk.startswith('data:'):
            chunk = chunk[5:].strip()
        if not chunk or chunk == '[DONE]':
            continue
        data = json.loads(chunk)
        assert 'choices' in data
        assert isinstance(data['choices'], list)
        assert len(data['choices']) == 1
        assert 'text' in data['choices'][0]
        assert isinstance(data['choices'][0]['text'], str)
        content += data['choices'][0]['text']
    assert content.startswith('The Zen of Python')

def run_complete_tests(client: TestClient):
    run_sync_complete_tests(client)
    run_stream_complete_tests(client)



import inspect, asyncio, typing

def tokens(str: str) -> int:
    return len(str)

async def asyncgen(
        func: typing.Callable,
        kvargs: dict,
    ) -> typing.AsyncGenerator[str, None]:

    '''Convert a chat/complete function to an async generator.'''
    sentinel = object()
    queue = asyncio.Queue()

    async def async_func():
        sig = inspect.signature(func)
        if 'callback' in sig.parameters:
            def callback(chunk: str):
                queue.put_nowait(chunk)
            kvargs['callback'] = callback

        result = func(**kvargs)
        if inspect.isasyncgen(result):
            async for chunk in result:
                queue.put_nowait(chunk)
        elif inspect.isgenerator(result):
            for chunk in result:
                queue.put_nowait(chunk)
        elif inspect.iscoroutine(result):
            queue.put_nowait(await result)
        else:
            queue.put_nowait(result)

    def sync_run():
        try:
            asyncio.run(async_func())
        finally:
            queue.put_nowait(sentinel)

    co = asyncio.to_thread(sync_run)
    asyncio.create_task(co)

    while True:
        chunk = await queue.get()
        if chunk is sentinel:
            break
        yield chunk



import uvicorn
import src.svllm as svllm

if __name__ == '__main__':
    svllm.set_chat(svllm.examples.chat)
    svllm.set_complete(svllm.examples.complete)
    svllm.set_embed(svllm.examples.embed)

    app = svllm.create_app()
    uvicorn.run(app, host='0.0.0.0', port=5261)



# https://platform.openai.com/docs/api-reference/introduction
# https://github.com/lm-sys/FastChat/blob/df817986a723fb513060e48cec1225a84110605d/fastchat/protocol/openai_api_protocol.py
from typing import Literal, Optional, List, Dict, Any, Union

import time, shortuuid
from pydantic import BaseModel, Field, model_serializer

class ErrorDetail(BaseModel):
    type: str
    message: str
    param: Optional[str] = None
    code: Optional[str] = None


class ErrorResponse(BaseModel):
    error: ErrorDetail


class ModelCard(BaseModel):
    id: str
    object: str = 'model'
    created: int = Field(default_factory=lambda: int(time.time()))
    owned_by: str = 'svllm'


class ModelList(BaseModel):
    object: str = 'list'
    data: List[ModelCard] = []


class UsageInfo(BaseModel):
    prompt_tokens: int = 0
    total_tokens: int = 0
    completion_tokens: Optional[int] = 0


class LogProbs(BaseModel):
    text_offset: List[int] = Field(default_factory=list)
    token_logprobs: List[Optional[float]] = Field(default_factory=list)
    tokens: List[str] = Field(default_factory=list)
    top_logprobs: List[Optional[Dict[str, float]]] = Field(default_factory=list)


class ChatCompletionRequest(BaseModel):
    model: str
    messages: Union[
        str,
        List[Dict[str, str]],
        List[Dict[str, Union[str, List[Dict[str, Union[str, Dict[str, str]]]]]]],
    ]
    temperature: Optional[float] = 0.7
    top_p: Optional[float] = 1.0
    top_k: Optional[int] = -1
    n: Optional[int] = 1
    max_tokens: Optional[int] = None
    stop: Optional[Union[str, List[str]]] = None
    stream: Optional[bool] = False
    presence_penalty: Optional[float] = 0.0
    frequency_penalty: Optional[float] = 0.0
    user: Optional[str] = None


class ChatMessage(BaseModel):
    role: str
    content: str


class ChatCompletionResponseChoice(BaseModel):
    index: int
    message: ChatMessage
    finish_reason: Optional[Literal['stop', 'length']] = None


class ChatCompletionResponse(BaseModel):
    id: str = Field(default_factory=lambda: f'chatcmpl-{shortuuid.random()}')
    object: str = 'chat.completion'
    created: int = Field(default_factory=lambda: int(time.time()))
    model: str
    choices: List[ChatCompletionResponseChoice]
    usage: UsageInfo


class DeltaMessage(BaseModel):
    role: Optional[str] = None
    content: Optional[str] = None

    @model_serializer
    def _serialize(self):
        return {k: v for k, v in self if v is not None}


class ChatCompletionResponseStreamChoice(BaseModel):
    index: int
    delta: DeltaMessage
    finish_reason: Optional[Literal['stop', 'length']] = None


class ChatCompletionStreamResponse(BaseModel):
    id: str = Field(default_factory=lambda: f'chatcmpl-{shortuuid.random()}')
    object: str = 'chat.completion.chunk'
    created: int = Field(default_factory=lambda: int(time.time()))
    model: str
    choices: List[ChatCompletionResponseStreamChoice]


class TokenCheckRequestItem(BaseModel):
    model: str
    prompt: str
    max_tokens: int


class TokenCheckRequest(BaseModel):
    prompts: List[TokenCheckRequestItem]


class TokenCheckResponseItem(BaseModel):
    fits: bool
    tokenCount: int
    contextLength: int


class TokenCheckResponse(BaseModel):
    prompts: List[TokenCheckResponseItem]


class EmbeddingsRequest(BaseModel):
    model: Optional[str] = None
    engine: Optional[str] = None
    input: Union[str, List[str]]
    user: Optional[str] = None
    encoding_format: Optional[str] = None

class Embedding(BaseModel):
    object: str = 'embedding'
    embedding: List[float]
    index: int


class EmbeddingUsageInfo(BaseModel):
    prompt_tokens: int = 0
    total_tokens: int = 0


class EmbeddingsResponse(BaseModel):
    object: str = 'list'
    data: List[Embedding]
    model: str
    usage: EmbeddingUsageInfo


class CompletionRequest(BaseModel):
    model: str
    prompt: Union[str, List[Any]]
    suffix: Optional[str] = None
    temperature: Optional[float] = 0.7
    n: Optional[int] = 1
    max_tokens: Optional[int] = 16
    stop: Optional[Union[str, List[str]]] = None
    stream: Optional[bool] = False
    top_p: Optional[float] = 1.0
    top_k: Optional[int] = -1
    logprobs: Optional[int] = None
    echo: Optional[bool] = False
    presence_penalty: Optional[float] = 0.0
    frequency_penalty: Optional[float] = 0.0
    user: Optional[str] = None
    best_of: Optional[int] = None


class CompletionResponseChoice(BaseModel):
    index: int
    text: str
    logprobs: Optional[LogProbs] = None
    finish_reason: Optional[Literal['stop', 'length']] = None


class CompletionResponse(BaseModel):
    id: str = Field(default_factory=lambda: f'cmpl-{shortuuid.random()}')
    object: str = 'text_completion'
    created: int = Field(default_factory=lambda: int(time.time()))
    model: str
    choices: List[CompletionResponseChoice]
    usage: UsageInfo


class CompletionResponseStreamChoice(BaseModel):
    index: int
    text: str
    logprobs: Optional[LogProbs] = None
    finish_reason: Optional[Literal['stop', 'length']] = None


class CompletionStreamResponse(BaseModel):
    id: str = Field(default_factory=lambda: f'cmpl-{shortuuid.random()}')
    object: str = 'text_completion'
    created: int = Field(default_factory=lambda: int(time.time()))
    model: str
    choices: List[CompletionResponseStreamChoice]



#!/usr/bin/env python3

import sys, argparse, importlib, svllm, uvicorn
from termcolor import colored

def import_from(location: str, name: str):
    module_str, _, attrs_str = location.partition(':')
    if not module_str:
        raise ValueError(f'{name} must be in the format \'module:attr\'.')
    instance = importlib.import_module(module_str)
    for attr in (attrs_str or name).split('.'):
        instance = getattr(instance, attr, None)
    if not instance:
        raise ImportError(f'\'{attrs_str or name}\' not found in \'{module_str}\'')
    return (instance, f'{module_str}:{attrs_str or name}')

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--chat', type=str, required=False, help='chat function')
    parser.add_argument('--complete', type=str, required=False, help='complete function')
    parser.add_argument('--embed', type=str, required=False, help='embed function')

    parser.add_argument('--prefix', type=str, default='/v1', required=False, help='api prefix')
    parser.add_argument('--host', type=str, default='127.0.0.1', required=False, help='host address')
    parser.add_argument('--port', type=int, default=5261, required=False, help='port number')
    parser.add_argument('--quiet', action='store_true', help='suppress output')

    cmd_args = parser.parse_args()
    sys.path.insert(0, '.')

    def log(message: str, error = False):
        if not cmd_args.quiet:
            color: str = 'red' if error else 'green'
            print(colored('SVLLM', color) + f':    {message}')

    if cmd_args.chat:
        chat, location = import_from(cmd_args.chat, 'chat')
        svllm.base.set_chat(chat)
        log(f'Chat function set to ' + colored(location, 'yellow'))

    if cmd_args.complete:
        complete, location = import_from(cmd_args.complete, 'complete')
        svllm.base.set_complete(complete)
        log(f'Complete function set to ' + colored(location, 'yellow'))

    if cmd_args.embed:
        embed, location = import_from(cmd_args.embed, 'embed')
        svllm.base.set_embed(embed)
        log(f'Embed function set to ' + colored(location, 'yellow'))

    log_level = 'warning' if cmd_args.quiet else 'info'
    app = svllm.create_app(prefix=cmd_args.prefix)
    uvicorn.run(app, host=cmd_args.host, port=cmd_args.port, log_level=log_level)

if __name__ == "__main__":
    main()



from .protocol import Chat, Complete, Embed, Model, ModelInfo
from typing import Optional, List

''' chat '''

_chat: Optional[Chat] = None

def set_chat(chat: Chat) -> Chat:
    '''Set the chat function.'''
    global _chat
    _chat = chat
    return

def get_chat() -> Optional[Chat]:
    '''Get the chat function.'''
    return _chat

''' complete '''

_complete: Optional[Complete] = None

def set_complete(complete: Complete) -> Complete:
    '''Set the complete function.'''
    global _complete
    _complete = complete
    return _complete

def get_complete() -> Optional[Complete]:
    '''Get the complete function.'''
    return _complete

''' embed '''

_embed: Optional[Embed] = None

def set_embed(embed: Embed) -> Embed:
    '''Set the embed function.'''
    global _embed
    _embed = embed
    return _embed

def get_embed() -> Optional[Embed]:
    '''Get the embed function.'''
    return _embed

''' models '''

_models: Optional[List[Model]] = None

def set_models(models: List[Model]) -> List[Model]:
    '''Set the list of models.'''
    global _models
    _models = models
    return _models

def get_models() -> List[Model]:
    '''Get the list of models.'''
    if _models is None:
        def _create_model(base: Optional[ModelInfo]):
            if base is None:
                return None
            model_id = getattr(base, '__model__', base.__name__)
            owned_by = getattr(base, '__owner__', 'svllm')
            return Model({ 'id': model_id, 'owned_by': owned_by })

        models = [
            _create_model(get_chat()),
            _create_model(get_complete()),
            _create_model(get_embed()),
        ]
        return [m for m in models if m is not None]
    return _models or []
